/// /////////////////////////////////////////////////////////////
/// the following flags affect the way the stream sink works
/// /////////////////////////////////////////////////////////////

/// define one or the other to use the cached buffer allocator
/// or the allocator of the stock file sink, otherwise the base
/// buffer allocator is used
//#define USE_SMART_SINK_ALLOCATOR
//#define USE_WRITER_SINK_ALLOCATOR

/// /////////////////////////////////////////////////////////////
/// the following flags only affect the test code in main()
/// /////////////////////////////////////////////////////////////

/// define one or the other to enable the code to test the stream
/// sink or the output pin, otherwise the stock file sink is used
//#define TEST_STREAM_SINK
//#define TEST_OUTPUT_PIN

/// enables the code to test the IStreamBuilder interface on the output pin
/// ignored unless TEST_OUTPUT_PIN is defined
//#define TEST_OUTPUT_BUILDER


#define _CRT_NON_CONFORMING_SWPRINTFS
#define _CRT_SECURE_NO_WARNINGS
#include "wmvmux.h"
#include "utils.h"
#include <nserror.h>
#pragma comment(lib,"wmvcore")
#include <mmreg.h>
#include <stdio.h>


#define DECLARE_IUNKNOWN2(base) \
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppv) { return base::GetOwner()->QueryInterface(riid,ppv); } \
	ULONG STDMETHODCALLTYPE AddRef() { return base::GetOwner()->AddRef(); } \
	ULONG STDMETHODCALLTYPE Release() { return base::GetOwner()->Release(); }


class __declspec(uuid("{6B6D0801-9ADA-11D0-A520-00A0D10129C0}")) _MEDIASUBTYPE_ASF;
#define MEDIASUBTYPE_ASF __uuidof(_MEDIASUBTYPE_ASF)


#ifndef USE_WRITER_SINK_ALLOCATOR

class CAsfMuxBuffer : public CUnknown, public INSSBuffer
{
private:
	DWORD	cb, n;
	LPBYTE	pb;

protected:
	DWORD	size(void) { return cb; }
	DWORD	used(void) { return n; }
	LPBYTE	data(void) { return pb; }

public:
	CAsfMuxBuffer(HRESULT* phr, IUnknown* pOuter, DWORD cb)
		: CUnknown(TEXT("AsfMuxBuffer"),pOuter,phr)
		, cb(0), n(0), pb(NULL)
	{
		if(S_OK != *phr) return;
		if(cb < 1) { *phr = E_INVALIDARG; return; }
		if(NULL == (this->pb = new BYTE[cb])) { *phr = E_OUTOFMEMORY; return; }
		this->cb = cb;
	}
	virtual ~CAsfMuxBuffer(void)
	{
		if(pb) delete [] pb;
	}

	static HRESULT STDMETHODCALLTYPE CreateInstance(DWORD cb, INSSBuffer** ppv)
	{
		if(ppv == NULL) return E_POINTER; else *ppv = NULL;
		HRESULT hr = S_OK; CAsfMuxBuffer* p = new CAsfMuxBuffer(&hr,NULL,cb);
		if(p == NULL) return E_OUTOFMEMORY; if(hr != S_OK) return hr;
		(*ppv = (INSSBuffer*)p)->AddRef(); return S_OK;
	}

	/// *** IUnknown *** ///
public:
	DECLARE_IUNKNOWN

	/// *** INonDelegatingUnknown *** ///
public:
	HRESULT STDMETHODCALLTYPE NonDelegatingQueryInterface(REFIID riid, void** ppv)
	{
		if(riid == __uuidof(INSSBuffer)) return GetInterface((INSSBuffer*)this,ppv);
		return CUnknown::NonDelegatingQueryInterface(riid,ppv);
	}

	/// *** INSSBuffer *** ///
public:
	HRESULT STDMETHODCALLTYPE GetLength(DWORD* pdwLength)
	{
		return pdwLength == NULL ? E_POINTER : ((*pdwLength = n), S_OK);
	}
	HRESULT STDMETHODCALLTYPE SetLength(DWORD dwLength)
	{
		return dwLength > cb ? E_INVALIDARG : ((n = dwLength), S_OK);
	}
	HRESULT STDMETHODCALLTYPE GetMaxLength(DWORD* pdwLength)
	{
		return pdwLength == NULL ? E_POINTER : ((*pdwLength = cb), S_OK);
	}
	HRESULT STDMETHODCALLTYPE GetBuffer(LPBYTE* ppbBuffer)
	{
		return ppbBuffer == NULL ? E_POINTER : ((*ppbBuffer = pb), S_OK);
	}
	HRESULT STDMETHODCALLTYPE GetBufferAndLength(LPBYTE* ppbBuffer, DWORD* pdwLength)
	{
		return ppbBuffer == NULL || pdwLength == NULL ? E_POINTER : ((*ppbBuffer = pb), (*pdwLength = n), S_OK);
	}
};

#endif /// !USE_WRITER_SINK_ALLOCATOR

#ifdef USE_SMART_SINK_ALLOCATOR

#define PTR2DWORD(x) ((DWORD)(DWORD_PTR)(x))
#define DWORD2PTR(x) ((LPVOID)(DWORD_PTR)(x))


class CBaseObjectCache
{
private:
	CRITICAL_SECTION		m_csLock;
	bool					m_zLock;
	HANDLE					m_hEvent;
	DWORD					m_dwSize;	/// 0 <= m_dwFree + m_dwUsed <= m_dwSize
	DWORD					m_dwFree;	/// 0 <= m_dwFree <= m_dwList
	DWORD					m_dwUsed;
	DWORD					m_dwList;
	LPVOID*					m_lpList;
	void (CALLBACK *m_fFreeCacheObject)(LPVOID obj);

protected:
	/** Create a new object that matches param. */
	virtual HRESULT STDMETHODCALLTYPE InitCacheObject(LPVOID param, LPVOID* pobj) = 0;
	/** @return S_OK if obj matches param, S_FALSE otherwise or an error */
	virtual HRESULT STDMETHODCALLTYPE SameCacheObject(LPVOID param, LPVOID obj) = 0;

	/** Same semantics as a mutex or critical section. */
	virtual HRESULT STDMETHODCALLTYPE LockObjectCache(void)
	{
		if(!m_zLock) return E_NOTIMPL;
		EnterCriticalSection(&m_csLock);
		return S_OK;
	}
	/** Same semantics as a mutex or critical section. */
	virtual void STDMETHODCALLTYPE FreeObjectCache(void)
	{
		if(!m_zLock) RaiseException(EXCEPTION_INVALID_HANDLE,0,0,NULL);
		LeaveCriticalSection(&m_csLock);
	}

	/** @return S_OK if signaled or S_FALSE on timeout */
	virtual HRESULT STDMETHODCALLTYPE WaitObjectCache(DWORD timeout)
	{
		if(!m_hEvent) return E_NOTIMPL;
		switch(WaitForSingleObject(m_hEvent,timeout))
		{
			case WAIT_OBJECT_0: return S_OK;
			case WAIT_TIMEOUT: return S_FALSE;
			case WAIT_FAILED: return HRESULT_FROM_WIN32(GetLastError());
			default: return E_UNEXPECTED;
		}
	}
	/** If flag, set signal otherwise reset signal. */
	virtual HRESULT STDMETHODCALLTYPE FlagObjectCache(bool flag)
	{
		if(!m_hEvent) return E_NOTIMPL;
		return (flag ? SetEvent(m_hEvent) : ResetEvent(m_hEvent)) ? S_OK : HRESULT_FROM_WIN32(GetLastError());
	}

public:
	/**
	 * @param dwMaxCacheSize can be INFINITE; return E_INVALIDARG if < 1
	 * @param fFreeCacheObject destroy the object; must be safe to call even after the derived class have been destroyed
	 * @param zUseLock if true, CBaseObjectCache will use an internal CRITICAL_SECTION as base implementation for Lock/FreeObjectCache()
	 * @param zUseFlag if true, CBaseObjectCache will use an internal event HANDLE as base implementation for Wait/FlagObjectCache()
	 */
	CBaseObjectCache(HRESULT* phr,
		void (CALLBACK *fFreeCacheObject)(LPVOID obj)
		, DWORD dwMaxCacheSize, bool zUseLock, bool zUseFlag)
		: m_zLock(false), m_hEvent(NULL)
		, m_dwSize(dwMaxCacheSize)
		, m_dwFree(0), m_dwUsed(0)
		, m_dwList(0), m_lpList(NULL)
		, m_fFreeCacheObject(fFreeCacheObject)
	{
		if(S_OK != *phr) return;

		HRESULT	hr		= S_OK;
		bool	zLock	= false;
		HANDLE	hEvent	= NULL;

		if(dwMaxCacheSize != INFINITE && dwMaxCacheSize < 1) { hr = E_INVALIDARG; goto exit; }
		if(fFreeCacheObject == NULL) { hr = E_INVALIDARG; goto exit; }

		if(zUseLock) {
			__try {
				InitializeCriticalSection(&m_csLock); zLock = true;
			} __except(GetExceptionCode() == STATUS_NO_MEMORY ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH) {
				hr = E_OUTOFMEMORY; goto exit;
			}
		}
		if(zUseFlag) {
			if(NULL == (hEvent = CreateEvent(NULL,FALSE,FALSE,NULL))) { hr = HRESULT_FROM_WIN32(GetLastError()); goto exit; }
		}

		m_zLock = zLock; zLock = false;
		m_hEvent = hEvent; hEvent = NULL;

	exit:
		if(zLock) DeleteCriticalSection(&m_csLock);
		if(hEvent) (void)CloseHandle(hEvent);
		*phr = hr;
	}
	virtual ~CBaseObjectCache(void)
	{
		if(m_lpList) {
			while(m_dwList-- > 0) if(m_lpList[m_dwList]) m_fFreeCacheObject(m_lpList[m_dwList]);
			delete [] m_lpList;
		}
		if(m_zLock) DeleteCriticalSection(&m_csLock);
		if(m_hEvent) (void)CloseHandle(m_hEvent);
	}

	/** @return S_OK or S_FALSE on timeout or an error (including E_POINTER if pobj is NULL); *pobj is NULL unless S_OK or E_POINTER */
	HRESULT STDMETHODCALLTYPE GetCacheObject(DWORD timeout, LPVOID param, LPVOID* pobj)
	{
		HRESULT hr		= S_OK;
		bool	lock	= false;

		if(pobj == NULL) { hr = E_POINTER; goto exit; } else *pobj = NULL;

		if(S_OK != (hr = LockObjectCache())) goto exit; else lock = true;

		for(;;) {
			for(DWORD i = 0, k = 0; k < m_dwFree && i < m_dwList; i++) {
				if(m_lpList[i]) {
					k++;
					switch(hr = SameCacheObject(param,m_lpList[i]))
					{
						case S_OK:
							if(S_OK == (hr = FlagObjectCache(false))) { *pobj = m_lpList[i]; m_dwUsed++; m_lpList[i] = NULL; m_dwFree--; }
							goto exit;
						case S_FALSE: continue;
						default: goto exit;
					}
				}
			}

			if(m_dwSize == INFINITE ||  m_dwFree + m_dwUsed < m_dwSize) {
				LPVOID obj = NULL;
				if(S_OK == (hr = InitCacheObject(param,&obj))) { *pobj = obj; m_dwUsed++; }
				goto exit;
			} else if(timeout != 0) {
				FreeObjectCache(); lock = false;
				DWORD time = timeout == INFINITE ? 0 : timeGetTime();
				switch(hr = WaitObjectCache(timeout))
				{
					case S_OK: break;
					case S_FALSE: goto exit;
					default: goto exit;
				}
				if(timeout != INFINITE) { time = timeGetTime() - time; timeout = timeout > time ? timeout - time : 0; }
				if(S_OK != (hr = LockObjectCache())) goto exit; else lock = true;
			} else {
				hr = S_FALSE; goto exit;
			}
		}

	exit:
		if(lock) FreeObjectCache();
		return hr;
	}
	/** @return S_OK or E_INVALIDARG if obj == NULL or another error */
	HRESULT STDMETHODCALLTYPE PutCacheObject(LPVOID obj)
	{
		HRESULT hr		= S_OK;
		bool	lock	= false;
		DWORD	k		= 0;

		if(obj == NULL) { hr = E_INVALIDARG; goto exit; }

		if(S_OK != (hr = LockObjectCache())) goto exit; else lock = true;

		if(m_dwFree < m_dwList) {
			for(; k < m_dwList; k++) if(m_lpList[k] == NULL) break;
		} else {
			LPVOID* p; DWORD n = m_dwList > 0 ? 2*m_dwList : 20; if(m_dwSize != INFINITE && n > m_dwSize) n = m_dwSize;
			if(NULL == (p = new LPVOID[n])) { hr = E_OUTOFMEMORY; goto exit; }
			if(m_lpList) {
				for(DWORD i = 0; k < m_dwFree && i < m_dwList; i++) if(m_lpList[i]) p[k++] = m_lpList[i];
				delete [] m_lpList;
			}
			m_dwList = n; m_lpList = p;
			for(DWORD i = k; i < m_dwList; i++) m_lpList[i] = NULL;
		}

		if(S_OK != (hr = FlagObjectCache(true))) goto exit;

		m_lpList[k] = obj; m_dwFree++; m_dwUsed--;

	exit:
		if(lock) FreeObjectCache();
		return hr;
	}
};

class CBaseCacheUnknown;

class CBaseUnknownCache : public CUnknown, private CBaseObjectCache
{
protected:
	virtual HRESULT STDMETHODCALLTYPE InitCacheUnknown(LPVOID param, CBaseCacheUnknown** pobj) = 0;
	virtual HRESULT STDMETHODCALLTYPE SameCacheUnknown(LPVOID param, CBaseCacheUnknown* obj) = 0;
	virtual HRESULT STDMETHODCALLTYPE LockUnknownCache(void) { return CBaseObjectCache::LockObjectCache(); }
	virtual void STDMETHODCALLTYPE FreeUnknownCache(void) { CBaseObjectCache::LockObjectCache(); }
	virtual HRESULT STDMETHODCALLTYPE WaitUnknownCache(DWORD timeout) { return CBaseObjectCache::WaitObjectCache(timeout); }
	virtual HRESULT STDMETHODCALLTYPE FlagUnknownCache(bool flag) { return CBaseObjectCache::FlagObjectCache(flag); }

protected:
	HRESULT STDMETHODCALLTYPE InitCacheObject(LPVOID param, LPVOID* pobj) { return InitCacheUnknown(param,(CBaseCacheUnknown**)pobj); }
	HRESULT STDMETHODCALLTYPE SameCacheObject(LPVOID param, LPVOID obj) { return SameCacheUnknown(param,(CBaseCacheUnknown*)obj); }
	HRESULT STDMETHODCALLTYPE LockObjectCache(void) { return LockUnknownCache(); }
	void STDMETHODCALLTYPE FreeObjectCache(void) { FreeUnknownCache(); }
	HRESULT STDMETHODCALLTYPE WaitObjectCache(DWORD dwTimeout) { return WaitUnknownCache(dwTimeout); }
	HRESULT STDMETHODCALLTYPE FlagObjectCache(bool flag) { return FlagUnknownCache(flag); }

private:
	static void STDMETHODCALLTYPE FreeCacheObject(LPVOID obj) { delete (CUnknown*)obj; }

public:
	CBaseUnknownCache(HRESULT* phr, LPCTSTR szDebugName, IUnknown* pOuter, DWORD dwMaxCacheSize, bool zUseLock, bool zUseFlag)
		: CUnknown(szDebugName,pOuter,phr)
		, CBaseObjectCache(phr,CBaseUnknownCache::FreeCacheObject,dwMaxCacheSize,zUseLock,zUseFlag)
	{
		if(S_OK != *phr) return;
	}
	virtual ~CBaseUnknownCache(void)
	{
	}

	/// *** IUnknown *** ///
public:
	DECLARE_IUNKNOWN

public:
	HRESULT STDMETHODCALLTYPE GetCacheUnknown(DWORD timeout, LPVOID param, REFIID riid, IUnknown** pobj)
	{
		HRESULT		hr	= S_OK;
		IUnknown*	obj	= NULL;
		if(pobj == NULL) { hr = E_POINTER; goto exit; } else *pobj = NULL;
		if(S_OK != (hr = GetCacheObject(timeout,param,(LPVOID*)&obj))) goto exit; else obj->AddRef();
		if(S_OK != (hr = obj->QueryInterface(riid,(void**)pobj))) goto exit;
	exit:
		if(obj) obj->Release();
		return hr;
	}
	HRESULT STDMETHODCALLTYPE PutCacheUnknown(IUnknown* obj)
	{
		return PutCacheObject(obj);
	}
};

class CBaseCacheUnknown : public CUnknown
{
private:
	LONG				m_lRef;
	CBaseUnknownCache*	m_pCache;

protected:
	ULONG STDMETHODCALLTYPE ReleaseCacheUnknown(void)
	{
		return CUnknown::NonDelegatingRelease();
	}

public:
	CBaseCacheUnknown(HRESULT* phr, LPCTSTR szDebugName, IUnknown* pOuter, CBaseUnknownCache* pCache)
		: CUnknown(szDebugName,pOuter,phr)
		, m_lRef(0), m_pCache(pCache)
	{
		if(S_OK != *phr) return;
		if(pCache == NULL) { *phr = E_INVALIDARG; return; }
	}

	/// *** IUnknown *** ///
public:
	DECLARE_IUNKNOWN

	/// *** INonDelegatingUnknown *** ///
public:
	ULONG STDMETHODCALLTYPE NonDelegatingAddRef(void)
	{
		LONG lRef = InterlockedIncrement(&m_lRef);
		if(lRef == 1) (void)m_pCache->NonDelegatingAddRef();
		return lRef;
	}
	ULONG STDMETHODCALLTYPE NonDelegatingRelease(void)
	{
		LONG lRef = InterlockedDecrement(&m_lRef);
		if(lRef > 0) return lRef;
		bool z = S_OK != m_pCache->PutCacheUnknown((IUnknown*)this);
		(void)m_pCache->NonDelegatingRelease();
		if(z) delete this;
		return 0UL;
	}
};


class CAsfMuxCacheBuffer : public CBaseCacheUnknown, public CAsfMuxBuffer
{
friend class CAsfMuxBufferCache;
public:
	CAsfMuxCacheBuffer(HRESULT* phr, IUnknown* pOuter, CBaseUnknownCache* pCache, DWORD cb)
		: CBaseCacheUnknown(phr,TEXT("AsfMuxCacheBuffer"),pOuter,pCache)
		, CAsfMuxBuffer(phr,pOuter,cb)
	{
	}

	/// *** IUnknown *** ///
public:
	DECLARE_IUNKNOWN2(CBaseCacheUnknown)

	/// *** INonDelegatingUnknown *** ///
public:
	HRESULT STDMETHODCALLTYPE NonDelegatingQueryInterface(REFIID riid, void** ppv)
	{
		if(S_OK == CAsfMuxBuffer::NonDelegatingQueryInterface(riid,ppv)) return S_OK;
		return CBaseCacheUnknown::NonDelegatingQueryInterface(riid,ppv);
	}
};

class CAsfMuxBufferCache : private CBaseUnknownCache, public IUnknown
{
public:
	CAsfMuxBufferCache(HRESULT* phr, IUnknown* pOuter, DWORD dwMaxCacheSize)
		: CBaseUnknownCache(phr,TEXT("AsfMuxBufferCache"),pOuter,dwMaxCacheSize,true,true)
	{
		if(S_OK != *phr) return;
	}

protected:
	HRESULT STDMETHODCALLTYPE InitCacheUnknown(LPVOID param, CBaseCacheUnknown** pobj)
	{
		if(pobj == NULL) return E_POINTER; else *pobj = NULL;
		HRESULT hr = S_OK; CAsfMuxCacheBuffer* p = new CAsfMuxCacheBuffer(&hr,NULL,this,PTR2DWORD(param));
		if(p == NULL) return E_OUTOFMEMORY;
		if(S_OK != hr) { delete p; return hr; }
		*pobj = p; return S_OK;
	}
	HRESULT STDMETHODCALLTYPE SameCacheUnknown(LPVOID param, CBaseCacheUnknown* obj)
	{
		return PTR2DWORD(param) <= ((CAsfMuxCacheBuffer*)obj)->size() ? S_OK : S_FALSE;
	}
	void STDMETHODCALLTYPE FreeCacheUnknown(CBaseCacheUnknown* obj)
	{
		delete obj;
	}

	/// *** IUnknown *** ///
public:
	DECLARE_IUNKNOWN

public:
	HRESULT STDMETHODCALLTYPE GetCacheBuffer(DWORD dwTimeout, DWORD dwBuffer, INSSBuffer** ppBuffer)
	{
		HRESULT		hr	= S_OK;
		if(ppBuffer == NULL) { hr = E_POINTER; goto exit; } else *ppBuffer = NULL;
		if(S_OK != (hr = GetCacheUnknown(dwTimeout,DWORD2PTR(dwBuffer),__uuidof(INSSBuffer),(IUnknown**)ppBuffer))) goto exit;
	exit:
		return hr;
	}
	HRESULT STDMETHODCALLTYPE PutCacheBuffer(INSSBuffer* pBuffer)
	{
		return PutCacheUnknown(pBuffer);
	}
};

#endif /// USE_SMART_SINK_ALLOCATOR


class CAsfMuxBaseSink : public CUnknown, public IWMWriterSink
{
private:
#if defined(USE_SMART_SINK_ALLOCATOR)
	CAsfMuxBufferCache*	m_pCache;
#elif defined(USE_WRITER_SINK_ALLOCATOR)
	IWMWriterSink*		m_pSink;
#endif
	ULONGLONG			m_llSize;

protected:
	CAsfMuxBaseSink(HRESULT* phr, LPCTSTR szDebugName)
		: CUnknown(szDebugName,NULL,phr)
#if defined(USE_SMART_SINK_ALLOCATOR)
		, m_pCache(NULL)
#elif defined(USE_WRITER_SINK_ALLOCATOR)
		, m_pSink(NULL)
#endif
		, m_llSize(0i64)
	{
		if(S_OK != *phr) return;

		HRESULT				hr		= S_OK;
#if defined(USE_SMART_SINK_ALLOCATOR)
		CAsfMuxBufferCache*	pCache	= NULL;
#elif defined(USE_WRITER_SINK_ALLOCATOR)
		IWMWriterSink*		pSink	= NULL;
#endif

#if defined(USE_SMART_SINK_ALLOCATOR)
		if(NULL == (pCache = new CAsfMuxBufferCache(&hr,NULL,INFINITE))) { hr = E_OUTOFMEMORY; goto exit; } else if(S_OK != hr) goto exit;
#elif defined(USE_WRITER_SINK_ALLOCATOR)
		if(S_OK != (hr = WMCreateWriterFileSink((IWMWriterFileSink**)&pSink))) goto exit;
#endif

#if defined(USE_SMART_SINK_ALLOCATOR)
		(m_pCache = pCache)->AddRef(); pCache = NULL;
#elif defined(USE_WRITER_SINK_ALLOCATOR)
		m_pSink = pSink; pSink = NULL;
#endif

#if defined(USE_SMART_SINK_ALLOCATOR)
	exit:
		if(pCache) delete pCache;
#elif defined(USE_WRITER_SINK_ALLOCATOR)
	exit:
		if(pSink) pSink->Release();
#endif
		*phr = hr;
	}
	virtual ~CAsfMuxBaseSink(void)
	{
#if defined(USE_SMART_SINK_ALLOCATOR)
		if(m_pCache) m_pCache->Release();
#elif defined(USE_WRITER_SINK_ALLOCATOR)
		if(m_pSink) m_pSink->Release();
#endif
	}

private:
	HRESULT STDMETHODCALLTYPE InternalWriteBuffer(bool head, INSSBuffer* pBuffer)
	{
		HRESULT	hr	= S_OK;
		LPBYTE	pb;
		DWORD	cb;

		if(S_OK != (hr = pBuffer->GetBufferAndLength(&pb,&cb))) goto exit;

		if(head && m_llSize > 0) {
			switch(hr = SeekSinkTarget(0Ui64))
			{
				case S_OK:
					if(S_OK != (hr = WriteSinkTarget(pBuffer,0i64,cb,pb))) goto exit;
					if(S_OK != (hr = SeekSinkTarget(m_llSize))) goto exit;
					break;
				case E_NOTIMPL: break;
				default: goto exit;
			}
		} else {
			if(S_OK != (hr = WriteSinkTarget(pBuffer,m_llSize,cb,pb))) goto exit; else m_llSize += cb;
		}

	exit:
		return hr;
	}

protected:
	virtual HRESULT STDMETHODCALLTYPE WriteSinkTarget(INSSBuffer* pBuffer, LONGLONG k, DWORD cb, LPBYTE pb) = 0;
	/** @return can return E_NOTIMPL */
	virtual HRESULT STDMETHODCALLTYPE SeekSinkTarget(LONGLONG k) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE CommitSinkTarget(LONGLONG k) { return S_OK; }

public:
	LONGLONG STDMETHODCALLTYPE GetSinkTargetSize(void) { return m_llSize; }

	/// *** IUnknown *** ///
public:
	DECLARE_IUNKNOWN

	/// *** INonDelegatingUnknown *** ///
public:
	HRESULT STDMETHODCALLTYPE NonDelegatingQueryInterface(REFIID riid, void** ppv)
	{
		if(riid == __uuidof(IWMWriterSink)) return GetInterface((IWMWriterSink*)this,ppv);
		return CUnknown::NonDelegatingQueryInterface(riid,ppv);
	}

	/// *** IWMWriterSink *** ///
public:
	HRESULT STDMETHODCALLTYPE OnHeader(INSSBuffer* pAsfHeader)
	{
		return InternalWriteBuffer(true,pAsfHeader);
	}
	HRESULT STDMETHODCALLTYPE IsRealTime(BOOL* pzRealTime)
	{
		if(pzRealTime == NULL) return E_POINTER; else { *pzRealTime = FALSE; return S_OK; }
	}
	HRESULT STDMETHODCALLTYPE AllocateDataUnit(DWORD cb, INSSBuffer** ppAsfBuffer)
	{
#if defined(USE_SMART_SINK_ALLOCATOR)
		return m_pCache->GetCacheBuffer(INFINITE,cb,ppAsfBuffer);
#elif defined(USE_WRITER_SINK_ALLOCATOR)
		return m_pSink->AllocateDataUnit(cb,ppAsfBuffer);
#else
		return CAsfMuxBuffer::CreateInstance(cb,ppAsfBuffer);
#endif
	}
	HRESULT STDMETHODCALLTYPE OnDataUnit(INSSBuffer* pAsfBuffer)
	{
		return InternalWriteBuffer(false,pAsfBuffer);
	}
	HRESULT STDMETHODCALLTYPE OnEndWriting(void)
	{
		LONGLONG ll = m_llSize; m_llSize = 0i64; return CommitSinkTarget(ll);
	}
};

class CAsfMuxStreamSink : public CAsfMuxBaseSink
{
private:
	IStream*			m_pStream;

	CAsfMuxStreamSink(HRESULT* phr, IStream* pStream)
		: CAsfMuxBaseSink(phr,TEXT("AsfMuxStreamSink"))
		, m_pStream(NULL)
	{
		if(S_OK != *phr) return;

		HRESULT				hr		= S_OK;

		if(pStream == NULL) { hr = E_INVALIDARG; goto exit; }

		(m_pStream = pStream)->AddRef();

	exit:
		*phr = hr;
	}
	~CAsfMuxStreamSink(void)
	{
		if(m_pStream) m_pStream->Release();
	}

public:
	static HRESULT STDMETHODCALLTYPE CreateInstance(IStream* pStream, IWMWriterSink** ppSink)
	{
		if(ppSink == NULL) return E_POINTER; else *ppSink = NULL;
		HRESULT hr = S_OK; CAsfMuxStreamSink* p = new CAsfMuxStreamSink(&hr,pStream);
		if(p == NULL) return E_OUTOFMEMORY; else if(S_OK != hr) return hr;
		(*ppSink = (IWMWriterSink*)p)->AddRef(); return S_OK;
	}

	/// *** CAsfMuxBaseSink *** ///
protected:
	HRESULT STDMETHODCALLTYPE WriteSinkTarget(INSSBuffer* pBuffer, LONGLONG k, DWORD cb, LPBYTE pb)
	{
		return m_pStream->Write(pb,cb,NULL);
	}
	HRESULT STDMETHODCALLTYPE SeekSinkTarget(LONGLONG k)
	{
		ULARGE_INTEGER i; LARGE_INTEGER d;
		return m_pStream->Seek(((d.QuadPart = k),d),STREAM_SEEK_SET,&i);
	}
	HRESULT STDMETHODCALLTYPE CommitSinkTarget(LONGLONG k)
	{
		return m_pStream->Commit(STGC_DEFAULT);
	}
};


class VAsfMuxFilter : public CBaseFilter
{
friend class CAsfMuxInputPin;
friend class CAsfMuxOutputPin;
friend class CAsfMuxAllocator;
friend class CAsfMuxSample;
protected:
	CCritSec			m_csLock;
	IWMProfile*			m_pAsfProfile;
	IWMWriter*			m_pAsfWriter;
	IWMWriterAdvanced3*	m_pAsfAdvanced3;
	bool				m_zRunning;

	VAsfMuxFilter(HRESULT* phr, IUnknown* pOuter)
		: CBaseFilter(NAME_AsfMuxFilter,pOuter,&m_csLock,CLSID_AsfMuxFilter,phr)
		, m_pAsfProfile(NULL), m_pAsfWriter(NULL), m_pAsfAdvanced3(NULL), m_zRunning(false)
	{
		if(S_OK != *phr) return;

		HRESULT				hr				= S_OK;
		IWMProfileManager*	pPrxManager		= NULL;
		IWMProfile*			pAsfProfile		= NULL;
		IWMWriter*			pAsfWriter		= NULL;
		IWMWriterAdvanced3*	pAsfAdvanced3	= NULL;

		if(S_OK != (hr = WMCreateProfileManager(&pPrxManager))) goto exit;
		if(S_OK != (hr = pPrxManager->CreateEmptyProfile(WMT_VER_9_0,&pAsfProfile))) goto exit;

		if(S_OK != (hr = WMCreateWriter(NULL,&pAsfWriter))) goto exit;
		if(S_OK != (hr = pAsfWriter->QueryInterface(__uuidof(IWMWriterAdvanced3),(void**)&pAsfAdvanced3))) goto exit;
		if(S_OK != (hr = pAsfAdvanced3->SetSyncTolerance(0))) goto exit;

		m_pAsfProfile	= pAsfProfile	; pAsfProfile	= NULL;
		m_pAsfWriter	= pAsfWriter	; pAsfWriter	= NULL;
		m_pAsfAdvanced3	= pAsfAdvanced3	; pAsfAdvanced3	= NULL;

	exit:
		if(pAsfAdvanced3) pAsfAdvanced3	->Release();
		if(pAsfWriter	) pAsfWriter	->Release();
		if(pAsfProfile	) pAsfProfile	->Release();
		if(pPrxManager	) pPrxManager	->Release();
		*phr = hr;
	}
	virtual ~VAsfMuxFilter(void)
	{
		if(m_pAsfAdvanced3	) m_pAsfAdvanced3	->Release();
		if(m_pAsfWriter		) m_pAsfWriter		->Release();
		if(m_pAsfProfile	) m_pAsfProfile		->Release();
	}

	virtual HRESULT STDMETHODCALLTYPE QueryInternalConnections(WORD wIndex, ULONG* pnPins, IPin** lpPins) = 0;
	virtual HRESULT STDMETHODCALLTYPE CheckMediaType(WORD wIndex, const AM_MEDIA_TYPE* pmt) = 0;
	virtual HRESULT STDMETHODCALLTYPE StartStreaming(WORD wIndex) = 0;
	virtual HRESULT STDMETHODCALLTYPE StopStreaming(WORD wIndex) = 0;
	virtual HRESULT STDMETHODCALLTYPE EndOfStream(WORD wIndex) = 0;
	virtual HRESULT STDMETHODCALLTYPE WriteStreamSample(WORD wIndex, IMediaSample* pms) = 0;
};


class CAsfMuxOutputPin : public CBaseOutputPin, public CAsfMuxBaseSink, public IStreamBuilder
{
friend class VAsfMuxFilter;
friend class CAsfMuxFilter;
private:
	VAsfMuxFilter*	m_pMux;
	bool			m_zActive;
	HRESULT			m_hrError;
	bool			m_zError;
	IStream*		m_pStream;
	BSTR			m_szFile;
	bool			m_zCreated;

	CAsfMuxOutputPin(HRESULT* phr, VAsfMuxFilter* pFilter, LPCWSTR szFile)
		: CBaseOutputPin(TEXT("AsfMuxOutputPin"),pFilter,pFilter->m_pLock,phr,L"out")
		, CAsfMuxBaseSink(phr,TEXT("CAsfMuxOutputPin"))
		, m_pMux(pFilter), m_zActive(false), m_hrError(S_OK), m_zError(true), m_pStream(NULL), m_szFile(NULL), m_zCreated(false)
	{
		if(S_OK != *phr) return;

		HRESULT	hr	= S_OK;
		BSTR	sz	= NULL;

		if(szFile) {
			if(NULL == (sz = SysAllocString(szFile))) { hr = E_OUTOFMEMORY; goto exit; }
		}

		m_szFile = sz; sz = NULL;

	exit:
		if(sz) SysFreeString(sz);
		*phr = hr;
	}
	virtual ~CAsfMuxOutputPin(void)
	{
		if(m_pStream) m_pStream->Release();
		if(m_szFile) SysFreeString(m_szFile);
	}

	/// *** IUnknown *** ///
public:
	DECLARE_IUNKNOWN2(CBaseOutputPin)

	/// *** INonDelegatingUnknown *** ///
public:
	HRESULT STDMETHODCALLTYPE NonDelegatingQueryInterface(REFIID riid, void** ppv)
	{
		if(m_szFile) {
			if(riid == __uuidof(IStreamBuilder)) return GetInterface((IStreamBuilder*)this,ppv);
		}
		if(S_OK == CAsfMuxBaseSink::NonDelegatingQueryInterface(riid,ppv)) return S_OK;
		return CBaseOutputPin::NonDelegatingQueryInterface(riid,ppv);
	}


	/// *** CAsfMuxBaseSink *** ///
protected:
	HRESULT STDMETHODCALLTYPE WriteSinkTarget(INSSBuffer* pBuffer, LONGLONG k, DWORD cb, LPBYTE pb)
	{
		HRESULT			hr	= S_OK;

		if(m_zError) return m_hrError;

		if(m_pStream == NULL) { hr = VFW_E_NOT_CONNECTED; goto exit; }

		if(S_OK != (hr = m_pStream->Write(pb,cb,NULL))) goto exit;

	exit:
		if(S_OK != hr) { m_hrError = hr; m_zError = true; }
		return hr;
	}

	/// *** CBasePin *** ///
public:
	HRESULT GetMediaType(int i, CMediaType* pmt)
	{
		if(i < 0) return E_INVALIDARG; else if(i > 0) return VFW_S_NO_MORE_ITEMS;
		pmt->SetType(&MEDIATYPE_Stream);
		pmt->SetSubtype(&MEDIASUBTYPE_ASF);
		return S_OK;
	}
	HRESULT CheckMediaType(const CMediaType* pmt)
	{
		if(pmt->majortype != MEDIATYPE_Stream) return S_FALSE;
		if(pmt->subtype != MEDIASUBTYPE_ASF) return S_FALSE;
		if(pmt->formattype != FORMAT_None && pmt->formattype != GUID_NULL) return VFW_E_INVALID_MEDIA_TYPE;
		if(pmt->cbFormat != 0 || pmt->pbFormat != NULL) return VFW_E_INVALID_MEDIA_TYPE;
		return S_OK;
	}

	HRESULT STDMETHODCALLTYPE Connect(IPin* pPin, const AM_MEDIA_TYPE* pmt)
	{
		HRESULT		hr		= S_OK;
		IStream*	pStream	= NULL;

		if(S_OK != (hr = pPin->QueryInterface(__uuidof(IStream),(void**)&pStream))) goto exit;
		if(S_OK != (hr = CBaseOutputPin::Connect(pPin,pmt))) goto exit;

		{ CAutoLock __auto_lock(m_pLock);
			if(m_pStream) m_pStream->Release(); m_pStream = pStream; pStream = NULL;
		} /// CAutoLock __auto_lock(m_pLock);

	exit:
		if(pStream) pStream->Release();
		return hr;
	}
	HRESULT STDMETHODCALLTYPE Disconnect(void)
	{
		HRESULT hr = S_OK;

		if(S_OK != (hr = CBaseOutputPin::Disconnect())) goto exit;

		{ CAutoLock __auto_lock(m_pLock);
			if(m_pStream) m_pStream->Release(); m_pStream = NULL;
		} /// CAutoLock __auto_lock(m_pLock);

	exit:
		return hr;
	}

	/// *** CBaseOutpuPin *** ///
public:
	HRESULT DecideBufferSize(IMemAllocator* pAlloc, ALLOCATOR_PROPERTIES* lpProps)
	{
		HRESULT hr = S_OK;
		ALLOCATOR_PROPERTIES apMin = { 1,64*1024,1,0 }, apReq = apMin, apRes = { 0,0,0,0 };
		if(lpProps->cBuffers > apReq.cBuffers) apReq.cBuffers = lpProps->cBuffers;
		if(lpProps->cbBuffer > apReq.cbBuffer) apReq.cbBuffer = lpProps->cbBuffer;
		if(lpProps->cbPrefix > apReq.cbPrefix) apReq.cbPrefix = lpProps->cbPrefix;
		if(S_OK != (hr = pAlloc->SetProperties(&apReq,&apRes))) goto exit;
		if(apRes.cBuffers < apMin.cBuffers) { hr = VFW_E_BUFFER_NOTSET; goto exit; }
		if(apRes.cbBuffer < apMin.cbBuffer) { hr = VFW_E_BUFFER_OVERFLOW; goto exit; }
		if(apRes.cbAlign != apMin.cbAlign ) { hr = VFW_E_BADALIGN; goto exit; }
		if(apRes.cbPrefix < apMin.cbPrefix) { hr = VFW_E_BADALIGN; goto exit; }
	exit:
		return hr;
	}
	HRESULT Active(void)
	{
		HRESULT hr = S_OK;
		if(S_OK != (hr = CBaseOutputPin::Active())) goto exit;
		m_hrError = S_OK; m_zError = false;
	exit:
		return hr;
	}

	/// *** IStreamBuilder *** ///
private:
	HRESULT STDMETHODCALLTYPE SetFileName(IUnknown* pFilter)
	{
		HRESULT				hr		= S_OK;
		IFileSinkFilter*	pSink	= NULL;
		IFileSinkFilter2*	pSink2	= NULL;
		LPOLESTR			szFile	= NULL;

		if(S_OK != (hr = pFilter->QueryInterface(__uuidof(IFileSinkFilter),(void**)&pSink))) goto exit;
		if(S_OK != (hr = pSink->GetCurFile(&szFile,NULL)) && E_FAIL != hr) goto exit;
		if(S_OK == hr && szFile != NULL && *szFile != L'\0') goto exit;
		if(S_OK != (hr = pSink->SetFileName(m_szFile,NULL))) goto exit;
		if(S_OK == pSink->QueryInterface(__uuidof(IFileSinkFilter2),(void**)&pSink2)) {
			(void)pSink2->SetMode(AM_FILE_OVERWRITE);
		}

	exit:
		if(szFile) CoTaskMemFree(szFile);
		if(pSink) pSink->Release();
		if(pSink2) pSink2->Release();
		return hr;
	}
public:
	HRESULT STDMETHODCALLTYPE Render(IPin* pThisPin, IGraphBuilder* pGrfBuilder)
	{
		HRESULT				hr			= S_OK;
		IEnumFilters*		pFilters	= NULL;
		IBaseFilter*		pFilter		= NULL;
		IFileSinkFilter*	pSink		= NULL;
		IEnumPins*			pPins		= NULL;
		IPin*				pPin		= NULL;
		IPin*				pPeer		= NULL;
		bool				zCreated	= false;

		if(pThisPin == NULL || pGrfBuilder == NULL) { hr = E_INVALIDARG; goto exit; }

		/// NOTE: this->IsConnected() == m_pStream != NULL

		if(S_OK != (hr = pGrfBuilder->EnumFilters(&pFilters))) goto exit;
		while(m_pStream == NULL && S_OK == (hr = pFilters->Next(1,&pFilter,NULL))) {
			/// TODO: skip myself!
			if(S_OK == pFilter->QueryInterface(__uuidof(IFileSinkFilter),(void**)&pSink)) {
				if(S_OK == pFilter->EnumPins(&pPins)) {
					while(m_pStream == NULL && S_OK == pPins->Next(1,&pPin,NULL)) {
						PIN_DIRECTION d;
						if(S_OK == pPin->QueryDirection(&d) && d == PINDIR_INPUT) {
							if(VFW_E_NOT_CONNECTED == (hr = pPin->ConnectedTo(&pPeer))) {
								if(S_OK == SetFileName(pSink)) {
									if(S_OK != pGrfBuilder->ConnectDirect(pThisPin,pPin,NULL)) {
										/// TODO: undo SetFileName()?
									}
								}
							} else if(S_OK == hr) { pPeer->Release(); pPeer = NULL; }
						}
						pPin->Release(); pPin = NULL;
					}
					pPins->Release(); pPins = NULL;
				}
				if(pSink) { pSink->Release(); pSink = NULL; }
			}
			pFilter->Release(); pFilter = NULL;
		} if(S_FALSE == hr) hr = S_OK; else goto exit;

		if(m_pStream == NULL) {
			if(S_OK != (hr = CoCreateInstance(CLSID_FileWriter,NULL,CLSCTX_INPROC_SERVER,__uuidof(IBaseFilter),(void**)&pFilter))) goto exit;
			if(S_OK != (hr = pGrfBuilder->AddFilter(pFilter,L"FileWriter"))) goto exit; else zCreated = true;
			if(S_OK != (hr = pFilter->FindPin(L"in",&pPin))) goto exit;
			if(S_OK != (hr = SetFileName(pFilter))) goto exit;
			if(S_OK != (hr = pGrfBuilder->ConnectDirect(pThisPin,pPin,NULL))) goto exit;
			m_zCreated = true; zCreated = false;
			pPin->Release(); pPin = NULL;
			pFilter->Release(); pFilter = NULL;
		}

	exit:
		if(pFilter && zCreated && pGrfBuilder) (void)pGrfBuilder->RemoveFilter(pFilter);
		if(pPeer	) pPeer		->Release();
		if(pPin		) pPin		->Release();
		if(pPins	) pPins		->Release();
		if(pFilter	) pFilter	->Release();
		if(pSink	) pSink		->Release();
		if(pFilters	) pFilters	->Release();
		return hr;
	}
	HRESULT STDMETHODCALLTYPE Backout(IPin* pThisPin, IGraphBuilder* pGrfBuilder)
	{
		HRESULT			hr			= S_OK;
		IPin*			pSinkPin	= NULL;
		PIN_INFO		pi			= { 0 };

		if(pThisPin == NULL || pGrfBuilder == NULL) { hr = E_INVALIDARG; goto exit; }

		if(S_OK != (hr = pThisPin->ConnectedTo(&pSinkPin))) { if(VFW_E_NOT_CONNECTED == hr) hr = S_OK; goto exit; }
		if(S_OK != (hr = pSinkPin->QueryPinInfo(&pi))) goto exit; else if(pi.pFilter == NULL) { hr = E_UNEXPECTED; goto exit; }
		if(S_OK != (hr = pGrfBuilder->Disconnect(pSinkPin))) goto exit;
		if(S_OK != (hr = pGrfBuilder->Disconnect(pThisPin))) goto exit;
		if(m_zCreated) {
			m_zCreated = false;
			if(S_OK != (hr = pGrfBuilder->RemoveFilter(pi.pFilter))) goto exit;
		}

	exit:
		if(pSinkPin		) pSinkPin		->Release();
		if(pi.pFilter	) pi.pFilter	->Release();
		return hr;
	}
};


class CAsfMuxInputPin : public CBaseInputPin
{
friend class VAsfMuxFilter;
friend class CAsfMuxFilter;
private:
	VAsfMuxFilter*	m_pMux;
	WORD			m_wIndex;
	LPWSTR			m_szBaseName;
	WCHAR			m_szThisName[16];
	bool			m_zIsAudio;
	bool			m_zEndOfStream;
	REFERENCE_TIME	m_rtLastTime;
	REFERENCE_TIME	m_rtNextTime;
	LONGLONG		m_rtTimePerSample;
	DWORD			m_dwSizePerSecond;
	HANDLE			m_hSleepEvent;
	bool			m_zSleeping;
	bool			m_zFirstSample;

	CAsfMuxInputPin(HRESULT* phr, VAsfMuxFilter* pFilter, WORD wIndex)
		: CBaseInputPin(TEXT("AsfMuxInputPin"),pFilter,pFilter->m_pLock,phr,L"in")
		, m_pMux(pFilter), m_wIndex(wIndex), m_szBaseName(NULL), m_zIsAudio(false), m_zEndOfStream(false)
		, m_rtLastTime(0), m_rtNextTime(-1), m_rtTimePerSample(0), m_dwSizePerSecond(0)
		, m_hSleepEvent(NULL), m_zSleeping(false), m_zFirstSample(true)
	{
		if(S_OK != *phr) return;

		HRESULT	hr			= S_OK;
		HANDLE	hSleepEvent	= NULL;

		(void)swprintf(m_szThisName,L"in/%02hu",m_wIndex);
		m_szBaseName = m_pName; m_pName = m_szThisName;

		if(NULL == (hSleepEvent = CreateEvent(NULL,FALSE,FALSE,NULL))) { hr = HRESULT_FROM_WIN32(GetLastError()); goto exit; }

		m_hSleepEvent = hSleepEvent; hSleepEvent = NULL;

	exit:
		if(hSleepEvent) (void)CloseHandle(hSleepEvent);
		*phr = hr;
	}
	virtual ~CAsfMuxInputPin(void)
	{
		if(m_hSleepEvent) (void)CloseHandle(m_hSleepEvent);
		m_pName = m_szBaseName;
	}

	/// *** CBasePin *** ///
public:
	HRESULT GetMediaType(int i, CMediaType* pmt)
	{
		return i < 0 ? E_INVALIDARG : VFW_S_NO_MORE_ITEMS;
	}
	HRESULT CheckMediaType(const CMediaType* pmt)
	{
		return m_pMux->CheckMediaType(m_wIndex,pmt);
	}
	HRESULT Active(void)
	{
		return m_pMux->StartStreaming(m_wIndex);
	}
	HRESULT Inactive(void)
	{
		return m_pMux->StopStreaming(m_wIndex);
	}
	HRESULT STDMETHODCALLTYPE EndOfStream(void)
	{
		return m_pMux->EndOfStream(m_wIndex);
	}
	HRESULT STDMETHODCALLTYPE QueryInternalConnections(IPin** lpPins, ULONG* pnPins)
	{
		return m_pMux->QueryInternalConnections(m_wIndex,pnPins,lpPins);
	}

	/// *** CBaseInputPin *** ///
public:
	HRESULT STDMETHODCALLTYPE GetAllocator(IMemAllocator** ppAllocator)
	{
		HRESULT hr = S_OK;
		CAutoLock __auto_lock(m_pLock);
		if(S_OK != (hr = CBaseInputPin::GetAllocator(ppAllocator))) goto exit;
	exit:
		return hr;
	}
HRESULT STDMETHODCALLTYPE NotifyAllocator(IMemAllocator* pAllocator, BOOL zReadOnly)
{
	return CBaseInputPin::NotifyAllocator(pAllocator,zReadOnly);
}
	HRESULT STDMETHODCALLTYPE Receive(IMediaSample* pms)
	{
		return m_pMux->WriteStreamSample(m_wIndex,pms);
	}
};


class CAsfMuxFilter : public VAsfMuxFilter, public IAsfMuxFilter, public IAMFilterMiscFlags, public IFileSinkFilter2
{
friend HRESULT STDMETHODCALLTYPE CreateInstance<CAsfMuxFilter,IBaseFilter>(IBaseFilter** ppv);
friend HRESULT STDAPICALLTYPE CreateAsfMuxFilter(IUnknown* pOuter, REFIID riid, void** ppv);
private:
	CAsfMuxOutputPin*	m_pOutput;
	WORD				m_wInputs;
	CAsfMuxInputPin*	m_pInputs[1+WM_MAX_STREAMS]; /// ASF stream numbers are in the range [1..63]
	bool				m_zShortest;
	BSTR				m_szFile;

	CAsfMuxFilter(HRESULT* phr, IUnknown* pOuter = NULL)
		: VAsfMuxFilter(phr,pOuter)
		, m_pOutput(NULL), m_wInputs(1), m_zShortest(false)
		, m_szFile(NULL)
	{
		if(S_OK != *phr) return;
	}
	~CAsfMuxFilter(void)
	{
		if(m_pOutput) {
			(void)m_pAsfAdvanced3->RemoveSink((IWMWriterSink*)m_pOutput);
			delete m_pOutput;
		}
		for(WORD wIndex = 1; wIndex < m_wInputs; wIndex++) delete m_pInputs[wIndex];
		if(m_szFile) SysFreeString(m_szFile);
	}

#if defined(_WINDLL)
public:
	static CUnknown* CALLBACK CreateInstance(IUnknown* pOuter, HRESULT* phr)
	{
		return new CAsfMuxFilter(phr,pOuter);
	}
#endif

private:
	HRESULT STDMETHODCALLTYPE SuspendInputStream(WORD wIndex)
	{
		HRESULT hr = S_OK;
		bool zSleep = false;
		{ CAutoLock __auto_lock(m_pLock);
			zSleep = !m_pInputs[wIndex]->m_zSleeping;
			if(zSleep) m_pInputs[wIndex]->m_zSleeping = true;
		} /// CAutoLock __auto_lock(m_pLock);
		if(zSleep) {
			switch(WaitForSingleObject(m_pInputs[wIndex]->m_hSleepEvent,INFINITE))
			{
				case WAIT_OBJECT_0: break;
				case WAIT_FAILED: hr = HRESULT_FROM_WIN32(GetLastError()); goto exit;
				case WAIT_TIMEOUT:
				default: hr = E_UNEXPECTED; goto exit;
			}
		}
	exit:
		{ CAutoLock __auto_lock(m_pLock);
			m_pInputs[wIndex]->m_zSleeping = false;
			(void)ResetEvent(m_pInputs[wIndex]->m_hSleepEvent);
		} /// CAutoLock __auto_lock(m_pLock);
		return hr;
	}
	HRESULT STDMETHODCALLTYPE ResumeInputStream(WORD wIndex)
	{
		{ CAutoLock __auto_lock(m_pLock);
			return m_pInputs[wIndex]->m_zSleeping ? (SetEvent(m_pInputs[wIndex]->m_hSleepEvent) ? S_OK : HRESULT_FROM_WIN32(GetLastError())) : S_OK;
		} /// CAutoLock __auto_lock(m_pLock);
	}

protected:
	HRESULT STDMETHODCALLTYPE QueryInternalConnections(WORD wIndex, ULONG* pnPins, IPin** lpPins)
	{
		ULONG nPins = 0;
		if(pnPins == NULL) return E_POINTER; else { nPins = *pnPins; *pnPins = 0; }
		if(nPins > 0 && lpPins == NULL) return E_INVALIDARG;
		{ CAutoLock __auto_lock(m_pLock);
			if(m_pOutput) {
				if(nPins < 1) return S_FALSE;
				*pnPins = 1; (lpPins[0] = m_pOutput)->AddRef();
			}
		} /// CAutoLock __auto_lock(m_pLock);
		return S_OK;
	}
	HRESULT STDMETHODCALLTYPE CheckMediaType(WORD wIndex, const AM_MEDIA_TYPE* pmt)
	{
		HRESULT				hr				= S_OK;
		IWMStreamConfig*	pAsfStream		= NULL;
		IWMMediaProps*		pAsfProps		= NULL;
		bool				zNewStream		= false;
		DWORD				dwBitRate		= 1000000000UL;
		LONGLONG			rtTimePerSample	= 0;
		DWORD				dwSizePerSecond	= 0;
		bool				zIsAudio		= false;
		CMediaType			mt;

		if(pmt->majortype == MEDIATYPE_Video) {
			/// == WMMEDIATYPE_Video
			if(pmt->formattype == FORMAT_VideoInfo) {
				VIDEOINFOHEADER* vih = (VIDEOINFOHEADER*)pmt->pbFormat;
				if(pmt->pbFormat == NULL || pmt->cbFormat < sizeof(*vih) || pmt->cbFormat < sizeof(*vih) - sizeof(vih->bmiHeader) + vih->bmiHeader.biSize) { hr = VFW_E_INVALID_MEDIA_TYPE; goto exit; }
				switch(vih->bmiHeader.biCompression)
				{
					case BI_RGB:
					case BI_BITFIELDS:
						break;
					case BI_RLE8:
					case BI_RLE4:
					case BI_JPEG:
					case BI_PNG:
						{ hr = VFW_E_TYPE_NOT_ACCEPTED; goto exit; }
					default:
						GUID g = MEDIATYPE_Video; g.Data1 = vih->bmiHeader.biCompression;
						if(g != pmt->subtype) { hr = VFW_E_TYPE_NOT_ACCEPTED; goto exit; }
				}
				if(S_OK != (hr = mt.Set(*pmt))) goto exit; else { pmt = &mt; vih = (VIDEOINFOHEADER*)pmt->pbFormat; }
				vih->dwBitRate = dwBitRate;
				rtTimePerSample = vih->AvgTimePerFrame;
			} else { hr = VFW_E_UNSUPPORTED_VIDEO; goto exit; }
		} else if(pmt->majortype == MEDIATYPE_Audio) {
			/// WMMEDIATYPE_Audio
			if(pmt->formattype == FORMAT_WaveFormatEx) {
				WAVEFORMATEX* wfx = (WAVEFORMATEX*)pmt->pbFormat;
				if(pmt->pbFormat == NULL || pmt->cbFormat < sizeof(*wfx) || pmt->cbFormat < sizeof(*wfx) + wfx->cbSize) { hr = VFW_E_INVALID_MEDIA_TYPE; goto exit; }
				if(wfx->wFormatTag == WAVE_FORMAT_EXTENSIBLE) {
					if(((WAVEFORMATEXTENSIBLE*)wfx)->SubFormat != pmt->subtype) { hr = VFW_E_TYPE_NOT_ACCEPTED; goto exit; }
				} else {
					GUID g = MEDIATYPE_Audio; g.Data1 = wfx->wFormatTag;
					if(pmt->subtype == MEDIASUBTYPE_MPEG1Packet || pmt->subtype == MEDIASUBTYPE_MPEG1Payload) {
						/// MPEG1WAVEFORMAT
						if(S_OK != (hr = mt.Set(*pmt))) goto exit; else { pmt = &mt; wfx = (WAVEFORMATEX*)pmt->pbFormat; }
						((AM_MEDIA_TYPE*)pmt)->subtype = g;
					} else if(pmt->subtype == MEDIASUBTYPE_DOLBY_AC3 || pmt->subtype == g && (
							   wfx->wFormatTag == WAVE_FORMAT_DVM		/// 0x2000
							|| wfx->wFormatTag == WAVE_FORMAT_RAW_SPORT	/// 0x0240
							|| wfx->wFormatTag == WAVE_FORMAT_ESST_AC3	/// 0x0241
						)) {
						/// WAVEFORMATEX: wFormatTag = WAVE_FORMAT_UNKNOWN
						if(S_OK != (hr = mt.Set(*pmt))) goto exit; else { pmt = &mt; wfx = (WAVEFORMATEX*)pmt->pbFormat; }
						g.Data1 = wfx->wFormatTag = WAVE_FORMAT_DOLBY_AC3_SPDIF;
						((AM_MEDIA_TYPE*)pmt)->subtype = g;
					} else if(pmt->subtype == g) {
						/// MEDIASUBTYPE_MPEG1AudioPayload == FOUCCMap(WAVE_FORMAT_MPEG)
						/// MEDIASUBTYPE_DOLBY_AC3_SPDIF == FOUCCMap(WAVE_FORMAT_DOLBY_AC3_SPDIF)
					} else { hr = VFW_E_TYPE_NOT_ACCEPTED; goto exit; }
				}
				if(wfx->wFormatTag == WAVE_FORMAT_MPEG) {
					MPEG1WAVEFORMAT* mpa = (MPEG1WAVEFORMAT*)wfx;
					if(mpa->fwHeadLayer == ACM_MPEG_LAYER3) {
						MPEGLAYER3WAVEFORMAT* mp3;
						if(NULL == (mp3 = (MPEGLAYER3WAVEFORMAT*)mt.AllocFormatBuffer((ULONG)sizeof(*mp3)))) { hr = E_OUTOFMEMORY; goto exit; } else { pmt = &mt; wfx = (WAVEFORMATEX*)pmt->pbFormat; }
						mp3->wfx				= mpa->wfx;
						mp3->wfx.wFormatTag		= WAVE_FORMAT_MPEGLAYER3;
						mp3->wfx.cbSize			= (WORD)(sizeof(*mp3) - sizeof(mp3->wfx));
						mp3->wID				= MPEGLAYER3_ID_MPEG;
						mp3->fdwFlags			= MPEGLAYER3_FLAG_PADDING_OFF;
						mp3->nBlockSize			= 500;
						mp3->nFramesPerBlock	= 1;
						mp3->nCodecDelay		= 0;
						((AM_MEDIA_TYPE*)pmt)->majortype	= MEDIATYPE_Audio;
						((AM_MEDIA_TYPE*)pmt)->subtype		= FOURCCMap(mp3->wfx.wFormatTag);
						((AM_MEDIA_TYPE*)pmt)->formattype	= FORMAT_WaveFormatEx;
					}
				}
				dwSizePerSecond = wfx->nAvgBytesPerSec;
				zIsAudio = true;
			} else { hr = VFW_E_UNSUPPORTED_AUDIO; goto exit; }
		} else if(pmt->majortype == WMMEDIATYPE_Image) {
			if(pmt->formattype == FORMAT_VideoInfo) {
				VIDEOINFOHEADER* vih = (VIDEOINFOHEADER*)pmt->pbFormat;
				if(pmt->pbFormat == NULL || pmt->cbFormat < sizeof(*vih)) { hr = VFW_E_INVALID_MEDIA_TYPE; goto exit; }
				if(S_OK != (hr = mt.Set(*pmt))) goto exit; else { pmt = &mt; vih = (VIDEOINFOHEADER*)pmt->pbFormat; }
				vih->dwBitRate = dwBitRate;
			} else { hr = VFW_E_UNSUPPORTED_STREAM; goto exit; }
		} else { hr = VFW_E_UNSUPPORTED_STREAM; goto exit; }

		{ CAutoLock __auto_lock(m_pLock);

		if(S_OK != (hr = m_pAsfProfile->GetStreamByNumber(wIndex,&pAsfStream))) {
			if(NS_E_NO_STREAM != hr) goto exit;
			if(S_OK != (hr = m_pAsfProfile->CreateNewStream(pmt->majortype,&pAsfStream))) goto exit; else zNewStream = true;
			if(S_OK != (hr = pAsfStream->SetStreamNumber(wIndex))) goto exit;
		}
		if(S_OK != (hr = pAsfStream->QueryInterface(__uuidof(IWMMediaProps),(void**)&pAsfProps))) goto exit;
		if(S_OK != (hr = pAsfProps->SetMediaType((WM_MEDIA_TYPE*)pmt))) goto exit;
		if(S_OK != (hr = pAsfStream->SetBufferWindow(-1))) goto exit;
		if(S_OK != (hr = pAsfStream->SetBitrate(dwBitRate))) goto exit;
		if(zNewStream) {
			if(S_OK != (hr = m_pAsfProfile->AddStream(pAsfStream))) goto exit;
		} else {
			if(S_OK != (hr = m_pAsfProfile->ReconfigStream(pAsfStream))) goto exit;
		}
		if(S_OK != (hr = m_pAsfWriter->SetProfile(m_pAsfProfile))) {
			if(NS_E_INVALID_STREAM == hr) hr = S_FALSE;
			if(zNewStream) {
				(void)m_pAsfProfile->RemoveStreamByNumber(wIndex);
			}
			goto exit;
		}

		m_pInputs[wIndex]->m_rtTimePerSample	= rtTimePerSample	;
		m_pInputs[wIndex]->m_dwSizePerSecond	= dwSizePerSecond	;
		m_pInputs[wIndex]->m_zIsAudio			= zIsAudio			;

		} /// CAutoLock __auto_lock(m_pLock);

	exit:
		if(pAsfProps	) pAsfProps	->Release();
		if(pAsfStream	) pAsfStream->Release();
		return hr;
	}
	HRESULT STDMETHODCALLTYPE StartStreaming(WORD wIndex)
	{
		HRESULT				hr				= S_OK;
		if(!m_zRunning) {
			{ CAutoLock __auto_lock(m_pLock);
				if(!m_zRunning) {
					DWORD n = 0;
					if(S_OK != (hr = m_pAsfWriter->GetInputCount(&n))) goto exit;
					for(DWORD i = 0; i < n; i++) {
						if(S_OK != (hr = m_pAsfWriter->SetInputProps(i,NULL))) goto exit;
					}
					if(S_OK != (hr = m_pAsfWriter->BeginWriting())) goto exit;
					m_pInputs[wIndex]->m_zFirstSample = true;
					m_zRunning = true;
				}
			} /// CAutoLock __auto_lock(m_pLock);
		}
	exit:
		return hr;
	}
	HRESULT STDMETHODCALLTYPE StopStreaming(WORD wIndex)
	{
		HRESULT hr = S_OK;
		{ CAutoLock __auto_lock(m_pLock);
			m_pInputs[wIndex]->m_zEndOfStream = false;
			if(m_zRunning) {
				if(S_OK != (hr = m_pAsfWriter->EndWriting())) goto exit;
				m_pInputs[wIndex]->m_zFirstSample = true;
				m_zRunning = false;
			}
		} /// CAutoLock __auto_lock(m_pLock);
	exit:
		return hr;
	}
	HRESULT STDMETHODCALLTYPE EndOfStream(WORD wIndex)
	{
		bool zRunning = false;
		CAutoLock __auto_lock(m_pLock);
		if(wIndex > 0) m_pInputs[wIndex]->m_zEndOfStream = true;
		for(WORD w = 1; w < m_wInputs; w++) {
			if(w == wIndex) continue;
			if(!m_pInputs[w]->IsConnected()) continue;
			if(m_pInputs[w]->m_zEndOfStream) continue;
			(void)ResumeInputStream(w);
			if(m_zShortest || wIndex == 0) {
				m_pInputs[w]->m_zEndOfStream = true;
			} else {
				zRunning = true;
			}
		}
		if(zRunning) return S_OK;
		if(m_pOutput) {
			if(m_pOutput->m_zError && m_pOutput->m_hrError != S_OK) {
				(void)m_pOutput->DeliverEndOfStream();
				(void)NotifyEvent(EC_ERRORABORT,m_pOutput->m_hrError,0);
				return m_pOutput->m_hrError;
			} else {
				return m_pOutput->DeliverEndOfStream();
			}
		} else {
			HRESULT hr = NotifyEvent(EC_COMPLETE,S_OK,NULL); if(S_FALSE == hr) hr = E_UNEXPECTED; return hr;
		}
	}
	HRESULT STDMETHODCALLTYPE WriteStreamSample(WORD wIndex, IMediaSample* pms)
	{
		HRESULT			hr			= S_OK;
		INSSBuffer*		pAsfBuffer	= NULL;
		bool			zSleep;

		if(S_OK != (hr = m_pInputs[wIndex]->CBaseInputPin::Receive(pms))) goto exit;

		LPBYTE pb1 = NULL, pb2 = NULL; LONG cb1 = -1L; DWORD cb2 = 0; REFERENCE_TIME rt1 = -1i64, rt2 = -1i64; DWORD dw = 0;
		if(S_OK != (hr = pms->GetPointer(&pb1))) goto exit; else if(pb1 == NULL) { hr = VFW_E_BUFFER_NOTSET; goto exit; }
		if((cb1 = pms->GetActualDataLength()) < 0) { hr = VFW_E_SIZENOTSET; goto exit; }
		switch(hr = pms->GetTime(&rt1,&rt2))
		{
			case S_OK: break;
			case VFW_S_NO_STOP_TIME: rt2 = -1i64; break;
			case VFW_E_SAMPLE_TIME_NOT_SET: rt1 = rt2 = -1i64; break;
			default: goto exit;
		}

		for(;;) {
			zSleep = false;
			{ CAutoLock __auto_lock(m_pLock);
				if(m_pInputs[wIndex]->m_zEndOfStream)
				{
					hr = S_FALSE;
					goto exit;
				}
				if(m_pOutput && m_pOutput->m_zError) {
					hr = EndOfStream(0);
					goto exit;
				}
				if(rt1 < 0) if((rt1 = m_pInputs[wIndex]->m_rtNextTime) < 0) { hr = VFW_E_SAMPLE_TIME_NOT_SET; goto exit; }
				if(rt2 < 0 || rt2 - rt1 <= 1) {
					if(m_pInputs[wIndex]->m_rtTimePerSample > 0) {
						rt2 = rt1 + m_pInputs[wIndex]->m_rtTimePerSample;
					} else if(m_pInputs[wIndex]->m_dwSizePerSecond > 0) {
						rt2 = rt1 + REFTIME_UNITS_PER_SEC * cb1 / m_pInputs[wIndex]->m_dwSizePerSecond;
					} else rt2 = -1i64;
				}
				m_pInputs[wIndex]->m_rtLastTime = rt1;
				m_pInputs[wIndex]->m_rtNextTime = rt2;
				for(WORD w = 1; w < m_wInputs; w++) {
					if(w == wIndex) continue;
					if(!m_pInputs[w]->IsConnected()) continue;
					if(m_pInputs[w]->m_zEndOfStream) continue;
					REFERENCE_TIME lag = rt1 - m_pInputs[w]->m_rtLastTime;
					REFERENCE_TIME max = m_pInputs[w]->m_zIsAudio && !m_pInputs[wIndex]->m_zIsAudio
						? -200 * REFTIME_UNITS_PER_MILLISEC
						: +500 * REFTIME_UNITS_PER_MILLISEC
					;
					if(lag > max) {
						(void)ResumeInputStream(w);
						zSleep = true;
					}
				}
			} /// CAutoLock __auto_lock(m_pLock);
			if(zSleep) { (void)SuspendInputStream(wIndex); } else { break; }
		}

		if(S_OK == pms->IsSyncPoint() || m_pInputs[wIndex]->m_zIsAudio) dw |= WM_SF_CLEANPOINT;
		if(S_OK == pms->IsDiscontinuity()) dw |= WM_SF_DISCONTINUITY;
		if(m_pInputs[wIndex]->m_zFirstSample) { dw |= WM_SF_DISCONTINUITY; m_pInputs[wIndex]->m_zFirstSample = false; }
		if(S_OK != (hr = m_pAsfWriter->AllocateSample((DWORD)cb1,&pAsfBuffer))) goto exit;
		if(S_OK != (hr = pAsfBuffer->GetBuffer(&pb2))) goto exit; if(pb2 == NULL) { hr = VFW_E_BUFFER_NOTSET; goto exit; }
		if(S_OK != (hr = pAsfBuffer->GetMaxLength(&cb2))) goto exit; if(cb2 < (DWORD)cb1) { hr = VFW_E_BUFFER_OVERFLOW; goto exit; }
		CopyMemory(pb2,pb1,(SIZE_T)cb1);
		if(S_OK != (hr = pAsfBuffer->SetLength((DWORD)cb1))) goto exit;
		if(S_OK != (hr = m_pAsfAdvanced3->WriteStreamSample(wIndex,(QWORD)rt1,0UL,0Ui64,dw,pAsfBuffer))) goto exit;

	exit:
		if(pAsfBuffer) { //CAutoLock __auto_lock(m_pLock);
			pAsfBuffer->Release();
		} /// CAutoLock __auto_lock(m_pLock);
		return hr;
	}

	/// *** IUnknown *** ///
public:
	DECLARE_IUNKNOWN

	/// *** INonDelegatingUnknown *** ///
public:
	HRESULT STDMETHODCALLTYPE NonDelegatingQueryInterface(REFIID riid, void** ppv)
	{
		if(riid == __uuidof(IAsfMuxFilter)) return GetInterface((IAsfMuxFilter*)this,ppv);
		if(riid == __uuidof(IAMFilterMiscFlags)) return GetInterface((IAMFilterMiscFlags*)this,ppv);
		if(m_pOutput == NULL) {
			if(riid == __uuidof(IFileSinkFilter2)) return GetInterface((IFileSinkFilter2*)this,ppv);
		}
		return VAsfMuxFilter::NonDelegatingQueryInterface(riid,ppv);
	}

	/// *** CBaseFilter *** ///
public:
	int GetPinCount()
	{
		HRESULT hr = S_OK;
		CAsfMuxInputPin* p = NULL;
		CAutoLock __auto_lock(m_pLock);
		WORD wCount = (WORD)(sizeof(m_pInputs) / sizeof(*m_pInputs));
		if(m_wInputs >= wCount) goto exit;
		for(WORD wIndex = 1; wIndex < m_wInputs; wIndex++) if(!m_pInputs[wIndex]->IsConnected()) goto exit;
		if(NULL != (p = new CAsfMuxInputPin(&hr,this,m_wInputs)) && S_OK == hr) {
			m_pInputs[m_wInputs++] = p; p = NULL;
			IncrementPinVersion();
		}
	exit:
		if(p) delete p;
		return (int)m_wInputs - (m_pOutput ? 0 : 1);
	}
	CBasePin* GetPin(int i)
	{
		CAutoLock __auto_lock(m_pLock);
		/// return output first so it gets activated and ready before inputs
		return m_pOutput
			? (i == 0 ? (CBasePin*)m_pOutput : i >= 1 && i < (int)m_wInputs ? (CBasePin*)m_pInputs[i] : NULL)
			: (++i >= 1 && i < (int)m_wInputs ? (CBasePin*)m_pInputs[i] : NULL);
	}

	/// *** IAMFilterMiscFlags *** ///
public:
	ULONG STDMETHODCALLTYPE GetMiscFlags(void)
	{
		CAutoLock __auto_lock(m_pLock);
		return m_pOutput ? 0 : AM_FILTER_MISC_FLAGS_IS_RENDERER;
	}

	/// *** IFileSinkFilter2 *** ///
public:
	HRESULT STDMETHODCALLTYPE SetFileName(LPCOLESTR szFileName, const AM_MEDIA_TYPE* pmt)
	{
		HRESULT	hr	= S_OK;
		BSTR	sz	= NULL;

		if(szFileName == NULL) { hr = E_INVALIDARG; goto exit; }

		if(pmt != NULL) {
			if(pmt->majortype != MEDIATYPE_Stream
			|| pmt->subtype != MEDIASUBTYPE_ASF
			|| pmt->formattype != FORMAT_None && pmt->formattype != GUID_NULL
			) { hr = VFW_E_TYPE_NOT_ACCEPTED; goto exit; }
			if(pmt->cbFormat != 0 || pmt->pbFormat != NULL) { hr = VFW_E_INVALID_MEDIA_TYPE; goto exit; }
		}

		if(NULL == (sz = SysAllocString(szFileName))) { hr = E_OUTOFMEMORY; goto exit; }

//		if(S_OK != (hr = this->EnableAsfMuxer(false,sz))) goto exit;
		if(S_OK != (hr = this->AddAsfFileSink(sz,TRUE,NULL))) goto exit;

		m_szFile = sz; sz = NULL;

	exit:
		if(sz) SysFreeString(sz);
		return hr;
	}
	HRESULT STDMETHODCALLTYPE GetCurFile(LPOLESTR* pszFileName, AM_MEDIA_TYPE* pmt)
	{
		HRESULT		hr	= S_OK;
		LPOLESTR	sz	= NULL;
		SIZE_T		n;

		if(pmt != NULL) {
			pmt->majortype				= MEDIATYPE_Stream;
			pmt->subtype				= MEDIASUBTYPE_ASF;
			pmt->bFixedSizeSamples		= FALSE;
			pmt->bTemporalCompression	= FALSE;
			pmt->lSampleSize			= 0;
			pmt->formattype				= FORMAT_None;
			pmt->pUnk					= NULL;
			pmt->cbFormat				= 0;
			pmt->pbFormat				= NULL;
		}

		if(pszFileName == NULL) { hr = E_POINTER; goto exit; } else *pszFileName = NULL;

		if(m_szFile == NULL) { hr = E_FAIL; goto exit; }

		n = sizeof(OLECHAR) * (SysStringLen(m_szFile) + 1);
		if(NULL == (sz = (LPOLESTR)CoTaskMemAlloc(n))) { hr = E_OUTOFMEMORY; goto exit; }
		CopyMemory(sz,m_szFile,n);

		*pszFileName = sz; sz = NULL;

	exit:
		if(sz) CoTaskMemFree(sz);
		return hr;
	}
	HRESULT STDMETHODCALLTYPE SetMode(DWORD dwFlags)
	{
		return dwFlags == AM_FILE_OVERWRITE ? S_OK : E_INVALIDARG;
	}
	HRESULT STDMETHODCALLTYPE GetMode(DWORD* pdwFlags)
	{
		return pdwFlags == NULL ? E_POINTER : ((*pdwFlags = AM_FILE_OVERWRITE),S_OK);
	}

	/// *** IAsfMuxFilter *** ///
public:
	void STDMETHODCALLTYPE StopAtShortestStream(BOOL enable)
	{
		CAutoLock __auto_lock(m_pLock);
		m_zShortest = enable != FALSE;
	}
	HRESULT STDMETHODCALLTYPE EnableAsfMuxer(BOOL enable, LPCWSTR szFile)
	{
		HRESULT hr = S_OK;
		CAsfMuxOutputPin* p = NULL;
		CAutoLock __auto_lock(m_pLock);
		if(enable) {
			if(m_pOutput != NULL) goto exit;
			DWORD count;
			if(S_OK != (hr = m_pAsfAdvanced3->GetSinkCount(&count))) goto exit;
			if(count > 0) { hr = VFW_E_WRONG_STATE; goto exit; }
			if(NULL == (p = new CAsfMuxOutputPin(&hr,this,szFile))) { hr = E_OUTOFMEMORY; goto exit; } if(S_OK != hr) goto exit;
			if(S_OK != (hr = m_pAsfAdvanced3->AddSink((IWMWriterSink*)p))) goto exit;
			m_pOutput = p; p = NULL;
		} else {
			if(m_pOutput == NULL) goto exit;
			if(m_pOutput->IsConnected()) { hr = VFW_E_WRONG_STATE; goto exit; }
			if(S_OK != (hr = m_pAsfAdvanced3->RemoveSink((IWMWriterSink*)m_pOutput))) goto exit;
			m_pOutput->Release(); m_pOutput = NULL;
		}
	exit:
		if(p) delete p;
		return hr;
	}
	HRESULT STDMETHODCALLTYPE AddAsfSink(IUnknown* pSink)
	{
		HRESULT			hr	= S_OK;
		IWMWriterSink*	p	= NULL;
		if(S_OK != (hr = pSink->QueryInterface(__uuidof(IWMWriterSink),(void**)&p))) goto exit;
		{ CAutoLock __auto_lock(m_pLock);
			if(m_pOutput != NULL) { hr = VFW_E_WRONG_STATE; goto exit; }
			if(S_OK != (hr = m_pAsfAdvanced3->AddSink(p))) goto exit;
		} /// CAutoLock __auto_lock(m_pLock);
	exit:
		if(p) p->Release();
		return hr;
	}
	HRESULT STDMETHODCALLTYPE AddAsfFileSink(LPCWSTR szFile, BOOL zDoAutoIndexing, IWMWriterFileSink3** ppSink)
	{
		HRESULT				hr		= S_OK;
		IWMWriterFileSink*	pSink	= NULL;
		IWMWriterFileSink3*	pSink3	= NULL;
		if(S_OK != (hr = WMCreateWriterFileSink(&pSink))) goto exit;
		if(S_OK != (hr = pSink->QueryInterface(__uuidof(IWMWriterFileSink3),(void**)&pSink3))) goto exit;
		if(S_OK != (hr = pSink3->Open(szFile))) goto exit;
		if(S_OK != (hr = pSink3->SetAutoIndexing(zDoAutoIndexing))) goto exit;
		if(S_OK != (hr = this->AddAsfSink(pSink3))) goto exit;
		if(ppSink) { *ppSink = pSink3; pSink3 = NULL; }
	exit:
		if(pSink	) pSink	->Release();
		if(pSink3	) pSink3->Release();
		return hr;
	}
	HRESULT STDMETHODCALLTYPE AddAsfNetworkSink(WORD wPort, WORD* pwPort, IWMWriterNetworkSink** ppSink)
	{
		HRESULT					hr		= S_OK;
		IWMWriterNetworkSink*	pSink	= NULL;
		DWORD					dwPort	= (DWORD)wPort;
		if(S_OK != (hr = WMCreateWriterNetworkSink(&pSink))) goto exit;
		if(S_OK != (hr = pSink->Open(&dwPort))) goto exit;
		if(S_OK != (hr = this->AddAsfSink(pSink))) goto exit;
		if(pwPort) *pwPort = (WORD)dwPort;
		if(ppSink) { *ppSink = pSink; pSink = NULL; }
	exit:
		if(pSink	) pSink	->Release();
		return hr;
	}
	HRESULT STDMETHODCALLTYPE AddAsfStreamSink(IStream* pStream, IWMWriterSink** ppSink)
	{
		HRESULT			hr		= S_OK;
		IWMWriterSink*	pSink	= NULL;
		if(S_OK != (hr = CAsfMuxStreamSink::CreateInstance(pStream,&pSink))) goto exit;
		if(S_OK != (hr = this->AddAsfSink(pSink))) goto exit;
		if(ppSink) { *ppSink = pSink; pSink = NULL; }
	exit:
		if(pSink	) pSink	->Release();
		return hr;
	}

	/**
	 * @param completed 0..100; always valid but updated only on WMT_INDEX_PROGRESS
	 * @return false to cancel indexing
	 */
	typedef bool (CALLBACK *tCreateAsfIndexCallback)(WMT_STATUS status, HRESULT hr, DWORD completed, LPVOID param);
	/**
	 * Always synchronous, even if fCallback is not NULL.
	 * @param wStream use 0 to index all streams
	 * @param dwInterval in frames if zFrameBased is true, millisecs otherwise; use 0 for default
	 * @param fCallback can be NULL
	 * @param pvParam ignored if fCallback is NULL
	 */
	static HRESULT STDMETHODCALLTYPE CreateAsfIndex(LPCWSTR szFile, WORD wStream, bool zFrameBased, DWORD dwInterval, tCreateAsfIndexCallback fCallback, LPVOID pvParam)
	{
		HRESULT			hr			= S_OK;
		IWMIndexer*		pAsfIndex	= NULL;
		IWMIndexer2*	pAsfIndex2	= NULL;
		class : public IWMStatusCallback {
		public:
			HANDLE					hEvent;
			tCreateAsfIndexCallback	fCallback;
			LPVOID					pvParam;
			bool					zCancel;
			DWORD					dwCompleted;
			int						iOpened, iStarted;
			HRESULT					hrError;
			void reset(tCreateAsfIndexCallback fCallback, LPVOID pvParam)
			{
				this->hEvent		= NULL;
				this->fCallback		= fCallback;
				this->pvParam		= pvParam;
				this->zCancel		= false;
				this->dwCompleted	= 0;
				this->iOpened		= 0;
				this->iStarted		= 0;
				this->hrError		= S_OK;
			}
			HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void** ppv)
			{
				if(ppv == NULL) return E_POINTER;
				if(riid == __uuidof(IUnknown) || riid == __uuidof(IWMStatusCallback)) {
					*ppv = (IWMStatusCallback*)this; return S_OK;
				} else {
					*ppv = NULL; return E_NOINTERFACE;
				}
			}
			ULONG STDMETHODCALLTYPE AddRef(void) { return 1UL; }
			ULONG STDMETHODCALLTYPE Release(void) { return 1UL; }
			HRESULT STDMETHODCALLTYPE OnStatus(WMT_STATUS status, HRESULT hr, WMT_ATTR_DATATYPE type, LPBYTE value, LPVOID context)
			{
				switch(status)
				{
					case WMT_ERROR			: break;
					case WMT_OPENED			: iOpened++; break;
					case WMT_STARTED		: iStarted++; break;
					case WMT_STOPPED		: iStarted--; break;
					case WMT_CLOSED			: iOpened--; break;
					case WMT_INDEX_PROGRESS	: if(type == WMT_TYPE_DWORD) dwCompleted = *(DWORD*)value; break;
				}
				if(hrError == S_OK && hr != S_OK) hrError = hr;
				if(fCallback) zCancel = !fCallback(status,hr,dwCompleted,pvParam);
				if(zCancel || iOpened <= 0 && iStarted <= 0) (void)SetEvent(hEvent);
				return S_OK;
			}
		}				scb;

		scb.reset(fCallback,pvParam);
		if(NULL == (scb.hEvent = CreateEvent(NULL,FALSE,FALSE,NULL))) { hr = HRESULT_FROM_WIN32(GetLastError()); goto exit; }
		if(S_OK != (hr = WMCreateIndexer(&pAsfIndex))) goto exit;
		if(S_OK != (hr = pAsfIndex->QueryInterface(__uuidof(IWMIndexer2),(void**)&pAsfIndex2))) goto exit;
		if(S_OK != (hr = pAsfIndex2->Configure(wStream,zFrameBased ? WMT_IT_FRAME_NUMBERS : WMT_IT_PRESENTATION_TIME,dwInterval == 0 ? NULL : &dwInterval,NULL))) goto exit;
		if(S_OK != (hr = pAsfIndex->StartIndexing(szFile,&scb,NULL))) goto exit;
		switch(WaitForSingleObject(scb.hEvent,INFINITE))
		{
			case WAIT_OBJECT_0: break;
			case WAIT_FAILED: hr = HRESULT_FROM_WIN32(GetLastError()); goto exit;
			default: hr = E_UNEXPECTED; goto exit;
		}
		if(scb.zCancel) {
			if(S_OK != (hr = pAsfIndex->Cancel())) goto exit;
		} else {
			hr = scb.hrError;
		}

	exit:
		if(pAsfIndex	) pAsfIndex		->Release();
		if(pAsfIndex2	) pAsfIndex2	->Release();
		if(scb.hEvent	) (void)CloseHandle(scb.hEvent);
		return hr;
	}
};


EXTERN_C HRESULT STDAPICALLTYPE CreateAsfMuxFilter(IUnknown* pOuter, REFIID riid, void** ppv)
{
	if(ppv == NULL) return E_POINTER; else *ppv = NULL;
	HRESULT hr = S_OK;
	CAsfMuxFilter* p = new CAsfMuxFilter(&hr,pOuter);
	if(NULL == p) { hr = E_OUTOFMEMORY; goto exit; }
	if(S_OK != hr) goto exit;
	if(S_OK != (hr = p->QueryInterface(riid,ppv))) goto exit;
	p = NULL;
exit:
	if(p) delete p;
	return hr;
}


#if defined(_CONSOLE)

enum { ARGPROG, ARG_OUT, ARG_SRC, ARGCOUNT };

int wmain(int argc, LPCWSTR argv[])
{
	HRESULT						hr				= S_OK;
	LONG						rc				= ERROR_SUCCESS;
	bool						com				= false;
	IGraphBuilder*				pGrfBuilder		= NULL;
	IObjectWithSite*			pGrfSite		= NULL;
	IMediaFilter*				pGrfFilter		= NULL;
	IMediaControl*				pGrfControl		= NULL;
	IAMGraphBuilderCallback*	pGrfCallback	= NULL;
	IBaseFilter*				pMuxFilter		= NULL;
	IAsfMuxFilter*				pMuxConfig		= NULL;
#if defined(TEST_STREAM_SINK)
	IStream*					pOutStream		= NULL;
#elif defined(TEST_OUTPUT_PIN)
	IBaseFilter*				pOutFilter		= NULL;
	IFileSinkFilter2*			pOutSink2		= NULL;
#endif

	if(argc < ARGCOUNT) {
		(void)printf("*** Usage: <out.wmv> <in.xyz> [<in.xyz>...]");
		hr = E_INVALIDARG; goto exit;
	}

	if(FAILED(hr = CoInitialize(NULL))) goto exit; else com = true;

	if(S_OK != (hr = CoCreateInstance(CLSID_FilterGraph,NULL,CLSCTX_INPROC_SERVER,__uuidof(IGraphBuilder),(void**)&pGrfBuilder))) goto exit;

	/// install our graph builder callback to reject unwanted filters
	if(S_OK != (hr = CreateInstance<CSomeDevelopersMustDieGraphBuilderCallback,IAMGraphBuilderCallback>(&pGrfCallback))) goto exit;
	if(S_OK != (hr = pGrfBuilder->QueryInterface(__uuidof(IObjectWithSite),(void**)&pGrfSite))) goto exit;
	if(S_OK != (hr = pGrfSite->SetSite(pGrfCallback))) goto exit; else { pGrfSite->Release(); pGrfSite = NULL; }
	/// the graph's IObjectWithSite::SetSite() does not AddRef() the callback, despite the IObjectWithSite contract,
	/// so the callback can only be safely Release()d after the graph is destroyed

	/// create our internal filter
	if(S_OK != (hr = CreateInstance<CAsfMuxFilter,IBaseFilter>(&pMuxFilter))) goto exit;
	if(S_OK != (hr = pGrfBuilder->AddFilter(pMuxFilter,L"AsfMuxFilter"))) goto exit;

	/// configure the ASF muxer
	if(S_OK != (hr = pMuxFilter->QueryInterface(__uuidof(IAsfMuxFilter),(void**)&pMuxConfig))) goto exit;
	pMuxConfig->StopAtShortestStream(TRUE);
#if defined(TEST_STREAM_SINK)
	if(S_OK != (hr = SHCreateStreamOnFile(argv[ARG_OUT],STGM_CREATE|STGM_DIRECT|STGM_READWRITE|STGM_SHARE_EXCLUSIVE,&pOutStream))) goto exit;
	if(S_OK != (hr = pMuxConfig->AddAsfStreamSink(pOutStream,NULL))) goto exit; else { pOutStream->Release(); pOutStream = NULL; }
#elif defined(TEST_OUTPUT_PIN)
#	ifdef TEST_OUTPUT_BUILDER
	if(S_OK != (hr = pMuxConfig->EnableAsfMuxer(TRUE,argv[ARG_OUT]))) goto exit;
#	else
	if(S_OK != (hr = pMuxConfig->EnableAsfMuxer(TRUE,NULL))) goto exit;
#	endif
#else
	if(S_OK != (hr = pMuxConfig->AddAsfFileSink(argv[ARG_OUT],TRUE,NULL))) goto exit;
#endif

	pMuxFilter->Release(); pMuxFilter = NULL;
	pMuxConfig->Release(); pMuxConfig = NULL;

#if defined(TEST_OUTPUT_PIN) && !defined(TEST_OUTPUT_BUILDER)
	if(S_OK != (hr = CoCreateInstance(CLSID_FileWriter,NULL,CLSCTX_INPROC_SERVER,__uuidof(IBaseFilter),(void**)&pOutFilter))) goto exit;
	if(S_OK != (hr = pGrfBuilder->AddFilter(pOutFilter,L"FileWriter"))) goto exit;
	if(S_OK != (hr = pOutFilter->QueryInterface(__uuidof(IFileSinkFilter2),(void**)&pOutSink2))) goto exit; else { pOutFilter->Release(); pOutFilter = NULL; }
	if(S_OK != (hr = pOutSink2->SetMode(AM_FILE_OVERWRITE))) goto exit;
	if(S_OK != (hr = pOutSink2->SetFileName(argv[ARG_OUT],NULL))) goto exit;
	pOutSink2->Release(); pOutSink2 = NULL;
#endif

	/// build the graph
	for(int argi = ARG_SRC; argi < argc; argi++) {
		LPCWSTR sz = argv[argi];
		if(S_OK != (hr = pGrfBuilder->RenderFile(argv[argi],NULL))) goto exit;
	}

	/// configure the graph to run as fast as possible
	if(S_OK != (hr = pGrfBuilder->QueryInterface(__uuidof(IMediaFilter),(void**)&pGrfFilter))) goto exit;
	if(S_OK != (hr = pGrfFilter->SetSyncSource(NULL))) goto exit; else { pGrfFilter->Release(); pGrfFilter = NULL; }

	/// run the graph
	if(S_OK != (hr = pGrfBuilder->QueryInterface(__uuidof(IMediaControl),(void**)&pGrfControl))) goto exit;
	if(S_OK != (hr = pGrfControl->Run()) && S_FALSE != hr) goto exit;
	while(S_OK != (hr = WaitForCompletion(pGrfBuilder,1000,NULL,NULL))) if(E_ABORT != hr) goto exit;
	pGrfControl->Release(); pGrfControl = NULL;

	pGrfBuilder->Release(); pGrfBuilder = NULL;
	pGrfCallback->Release(); pGrfCallback = NULL;

#if defined(TEST_STREAM_SINK) || defined(TEST_OUTPUT_PIN)
	/// index the ASF
	if(S_OK != (hr = CAsfMuxFilter::CreateAsfIndex(argv[ARG_OUT],0,true,0,NULL,NULL))) goto exit;
#endif

exit:
	if(pMuxConfig	) pMuxConfig	->Release();
#if defined(TEST_STREAM_SINK)
	if(pOutStream	) pOutStream	->Release();
#elif defined(TEST_OUTPUT_PIN)
	if(pOutSink2	) pOutSink2		->Release();
	if(pOutFilter	) pOutFilter	->Release();
#endif
	if(pMuxFilter	) pMuxFilter	->Release();
	if(pGrfControl	) pGrfControl	->Release();
	if(pGrfFilter	) pGrfFilter	->Release();
	if(pGrfSite		) pGrfSite		->Release();
	if(pGrfBuilder	) pGrfBuilder	->Release();
	if(pGrfCallback	) pGrfCallback	->Release();
	if(com			) CoUninitialize();
	return hr;
}

#elif defined(_WINDLL)

AMOVIESETUP_FILTER sudAsfMuxFilter = { &CLSID_AsfMuxFilter,NAME_AsfMuxFilter, MERIT_DO_NOT_USE, 0,NULL };
CFactoryTemplate g_Templates[] = {
	{ sudAsfMuxFilter.strName,sudAsfMuxFilter.clsID, CAsfMuxFilter::CreateInstance,NULL, &sudAsfMuxFilter },
};
int g_cTemplates = (int)(sizeof(g_Templates) / sizeof(*g_Templates));

EXTERN_C HRESULT STDAPICALLTYPE DllRegisterServer(void) { return AMovieDllRegisterServer2(TRUE); }
EXTERN_C HRESULT STDAPICALLTYPE DllUnregisterServer(void) { return AMovieDllRegisterServer2(FALSE); }

#elif defined (_LIB)

#elif
#	pragma error Only LIB, DLL or CONSOLE build configs are supported!
#endif

